@@ -3,7 +3,9 @@ use std::task::{Context, Poll};
 use std::{future::Future, marker::PhantomData, num::NonZeroU16, pin::Pin, rc::Rc};
 
 use ntex::service::{fn_factory_with_config, Service, ServiceFactory};
-use ntex::util::{inflight::InFlightService, join, Either, HashSet, Ready};
+use ntex::util::{
+    buffer::BufferService, inflight::InFlightService, join, Either, HashSet, Ready,
+};
 
 use crate::error::{MqttError, ProtocolError};
 use crate::io::DispatchItem;
@@ -51,11 +53,18 @@ where
         async move {
             let (publish, control) = fut.await;
 
+            let control = BufferService::new(
+                16,
+                || MqttError::<C::Error>::Disconnected,
+                // limit number of in-flight messages
+                InFlightService::new(1, control?.map_err(MqttError::Service)),
+            );
+
             Ok(
                 // limit number of in-flight messages
                 InFlightService::new(
                     inflight,
-                    Dispatcher::<_, _, _, E>::new(cfg, publish?, control?),
+                    Dispatcher::<_, _, _, E>::new(cfg, publish?, control),
                 ),
             )
         }
@@ -80,7 +89,7 @@ struct Inner<C> {
 impl<St, T, C, E> Dispatcher<St, T, C, E>
 where
     T: Service<Request = Publish, Response = (), Error = E>,
-    C: Service<Request = ControlMessage<E>, Response = ControlResult, Error = E>,
+    C: Service<Request = ControlMessage<E>, Response = ControlResult, Error = MqttError<E>>,
 {
     pub(crate) fn new(session: Session<St>, publish: T, control: C) -> Self {
         let sink = session.sink().clone();
@@ -98,7 +107,7 @@ where
 impl<St, T, C, E> Service for Dispatcher<St, T, C, E>
 where
     T: Service<Request = Publish, Response = (), Error = E>,
-    C: Service<Request = ControlMessage<E>, Response = ControlResult, Error = E>,
+    C: Service<Request = ControlMessage<E>, Response = ControlResult, Error = MqttError<E>>,
     C::Future: 'static,
     E: 'static,
 {
@@ -112,7 +121,7 @@ where
 
     fn poll_ready(&self, cx: &mut Context<'_>) -> Poll<Result<(), Self::Error>> {
         let res1 = self.publish.poll_ready(cx).map_err(MqttError::Service)?;
-        let res2 = self.inner.control.poll_ready(cx).map_err(MqttError::Service)?;
+        let res2 = self.inner.control.poll_ready(cx)?;
 
         if res1.is_pending() || res2.is_pending() {
             Poll::Pending
@@ -252,7 +261,7 @@ pin_project_lite::pin_project! {
 impl<T, C, E> Future for PublishResponse<T, C, E>
 where
     T: Service<Request = Publish, Response = (), Error = E>,
-    C: Service<Request = ControlMessage<E>, Response = ControlResult, Error = E>,
+    C: Service<Request = ControlMessage<E>, Response = ControlResult, Error = MqttError<E>>,
 {
     type Output = Result<Option<codec::Packet>, MqttError<E>>;
 
@@ -300,7 +309,7 @@ pin_project_lite::pin_project! {
 
 impl<C: Service, E> ControlResponse<C, E>
 where
-    C: Service<Request = ControlMessage<E>, Response = ControlResult, Error = E>,
+    C: Service<Request = ControlMessage<E>, Response = ControlResult, Error = MqttError<E>>,
 {
     #[allow(clippy::match_like_matches_macro)]
     fn new(pkt: ControlMessage<E>, inner: &Rc<Inner<C>>) -> Self {
@@ -315,7 +324,7 @@ where
 
 impl<C, E> Future for ControlResponse<C, E>
 where
-    C: Service<Request = ControlMessage<E>, Response = ControlResult, Error = E>,
+    C: Service<Request = ControlMessage<E>, Response = ControlResult, Error = MqttError<E>>,
 {
     type Output = Result<Option<codec::Packet>, MqttError<E>>;
 
@@ -350,13 +359,18 @@ where
             Poll::Ready(Err(err)) => {
                 // do not handle nested error
                 if *this.error {
-                    Poll::Ready(Err(MqttError::Service(err)))
+                    Poll::Ready(Err(err))
                 } else {
                     // handle error from control service
-                    *this.error = true;
-                    let fut = this.inner.control.call(ControlMessage::error(err));
-                    self.as_mut().project().fut.set(fut);
-                    self.poll(cx)
+                    match err {
+                        MqttError::Service(err) => {
+                            *this.error = true;
+                            let fut = this.inner.control.call(ControlMessage::error(err));
+                            self.as_mut().project().fut.set(fut);
+                            self.poll(cx)
+                        }
+                        _ => Poll::Ready(Err(err)),
+                    }
                 }
             }
             Poll::Pending => Poll::Pending,
